/* -------------------------------------------------------------
   file    : autodox/autodox-start.p
   by      : Murray Hermann
   purpose : Select some files to lint and launch autodox.p.
   
   Copied from prolint/start.p (Jurjen Dijkstra)
   ------------------------------------------------------------- */
{autodox/dlc-version.i}   
{autodox/sourcefiles.i}

define variable areasToScan as char                   no-undo.
DEFINE VARIABLE modalresult AS INTEGER   INITIAL 0    NO-UNDO.
DEFINE VARIABLE profile     AS CHARACTER              NO-UNDO.
DEFINE VARIABLE ClearOutput AS LOGICAL                NO-UNDO.
DEFINE VARIABLE autodox     AS HANDLE                 NO-UNDO.
DEFINE VARIABLE appName     AS CHARACTER              NO-UNDO.
DEFINE VARIABLE outDir      AS CHARACTER INITIAL "./" NO-UNDO.



/*TEMP-START*/
    assign outDir = "C:/htmlTest/"
           appName = "appName"
           areasToScan = "C:/".
/*TEMP-END*/

if session:batch-mode = TRUE then
do:
    run getBatchParams.
    if outDir = ? or appName = ? or areasToScan = ? then 
    do: 
        run sendMsg("Leaving autodox as incorrect parameters supplied.").
        quit.
    end.
end.

run getOutputDir.

IF modalresult=1 THEN 
do:
    RUN autodox/autodox3.p PERSISTENT SET autodox.

    run setProcHandle in autodox (this-procedure).
    RUN setOutputDir IN autodox (outDir).
    RUN setAppName IN autodox (appName).
    run setGenerateTableInfo IN autodox (TRUE). /*Generate HTML for DB tables*/
    
    RUN setProcCommentLoc IN autodox (true).    /* TRUE = Before, FALSE = After. See readme.html */
    RUN setFuncCommentLoc IN autodox (true).    /* TRUE = Before, FALSE = After. See readme.html */
    RUN setParamCommentLoc IN autodox (true).   /* TRUE = Before, FALSE = After. See readme.html */
    
    RUN main IN autodox.
    
    if session:batch-mode = FALSE then
        MESSAGE "Open" outDir + "index.html to view the generated documentation." VIEW-AS ALERT-BOX.
    
    APPLY "CLOSE":U TO autodox.
    autodox = ?.
end.

  /*Clean up*/
  &IF {&dlc-version}>8 &THEN
    empty temp-table tt_sourcefiles.
  &ELSE
    for each tt_sourcefiles: delete tt_sourcefiles. end.
  &ENDIF

quit.

PROCEDURE GetFirstLintSource :
  /* purpose: prolint.p calls this ip to ask for the first sourcefile to analyze.
              return ? if you don't have any sourcefiles. */
  DEFINE OUTPUT PARAMETER pSourceFile AS CHARACTER NO-UNDO.
  
  FIND FIRST tt_sourcefiles NO-ERROR.
  IF AVAILABLE tt_sourcefiles THEN 
     pSourceFile = tt_sourcefiles.SourceFile.
  ELSE 
     pSourceFile = ?.
  
END PROCEDURE.


PROCEDURE GetNextLintSource :
  /* purpose: prolint.p calls this ip to ask for the next sourcefile to analyze.
              return ? if you don't have any sourcefiles. */
  DEFINE OUTPUT PARAMETER pSourceFile AS CHARACTER NO-UNDO.
  
  FIND NEXT tt_sourcefiles NO-ERROR.
  IF AVAILABLE tt_sourcefiles THEN 
     pSourceFile = tt_sourcefiles.SourceFile.
  ELSE 
     pSourceFile = ?.
  
END PROCEDURE.
           

procedure getOutputDir :
    /* purpose: Get the name of the application and the output directory */
    REPEAT: 
    process events.

    RUN autodox/selectfiles.w 
        (input-output areasToScan, /*should be blank*/
         INPUT "Enter the name of the directory for the "
               + "output HTML files. File names in the file list should be "
               + "absolute. "
               + "Enter ./ for current (working) directory.",
         INPUT-OUTPUT outDir,
         INPUT "~n~n~n Enter your application's name.~n",
         INPUT-OUTPUT appName,
         OUTPUT modalresult, /* 1=ok, else is cancel */
         OUTPUT profile, 
         OUTPUT TABLE tt_sourcefiles,
         OUTPUT ClearOutput). 

      IF RETURN-VALUE = "cancel" THEN
         RETURN.
    
      IF LENGTH(outDir) > 0 
      AND SUBSTRING(outDir,LENGTH(outDir),1) <> "/"
      AND SUBSTRING(outDir,LENGTH(outDir),1) <> "~\"
      THEN DO: 
        if session:batch-mode = FALSE then
        do:
            MESSAGE "Please re-enter output directory name with a / or ~\ at the end." VIEW-AS ALERT-BOX. 
            UNDO, RETRY.
        end.
        else run sendMsg("Please re-enter output directory name with a / or ~\ at the end.").
      END. 
      ELSE LEAVE.
    END. 
end procedure.

procedure getBatchParams:
    /*purpose: Get the params from the -param startup option
    E.G.   ... 
-param outdir=D:/shared/api/htmlTest/,appName=myname,areasToScan=/tmp/!/a/ */

    def var c as char no-undo.
    def var i as int no-undo.

    assign outDir = ?
           appName = ?
           areasToScan = ?
           modalresult = 0.

    if session:parameter = ? or session:parameter = "" or
    not(session:parameter matches "*outdir*" and
        session:parameter matches "*appName*" and
        session:parameter matches "*areasToScan*"
        )then 
    do:
        run sendMsg("No parameters.  Call like: " +
'... -param outdir=D:/shared/api/htmlTest/,appName=myname,areasToScan=/tmp/!/a/ ...').
        return.
    end.

    repeat i = 1 to num-entries(session:parameter):
        c = entry(1, entry(i, session:parameter), "=").
        if c = "outdir" then
            outdir = entry(2, entry(i, session:parameter), "=").
        else if c = "appName" then
            appName = entry(2, entry(i, session:parameter), "=").
        else if c = "areasToScan" then
            areasToScan = replace(entry(2, entry(i, session:parameter), "="), "!", "~n").
    end.
end procedure.

procedure sendMsg:
    def input param msg as char no-undo.

    output to value(SESSION:TEMP-DIRECTORY + "joanju_autodox_err.txt") append unbuffered.
    put unformatted string(today, "99/99/9999") " " STRING(TIME, "HH:MM:SS") " " msg skip.
    output close.
end procedure.
